/*
** This file is part of "zebedee".
**
** Copyright 1999 by Neil Winton. All rights reserved.
** 
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** For further details on "zebedee" see http://www.winton.org.uk/zebedee/
**
*/

char *zebedee_c_rcsid = "$Id: zebedee.c,v 1.11 1999/11/07 16:01:49 nwinton Exp nwinton $";
#define RELEASE_STR "1.1.0"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>

#include "gmp.h"
#include "blowfish.h"
#include "zlib.h"
#ifndef DONT_HAVE_BZIP2
#include "bzlib.h"
#endif
#include "sha.h"

#ifdef __CYGWIN__
#undef WIN32
#endif

#ifdef WIN32
/*
** Windows-specific include files and macros
*/

#include <windows.h>
#include <io.h>
#include <winsock.h>
#include <process.h>
#include <mmsystem.h>

#include "getopt.h"

#define DFLT_SHELL	"c:\\winnt\\system32\\cmd.exe"
#define getpid()	GetCurrentProcessId()
#define FILE_SEP_CHAR	'\\'

/*
** Winsock state data
*/

static struct WSAData WsaState;

/*
** Global Mutex (aka critical section)
*/

CRITICAL_SECTION Mutex;

extern void svcRun(char *name);
extern int svcInstall(char *name, char *configFile);
extern int svcRemove(char *name);

#else /* !WIN32 */

#include <sys/types.h>
#include <sys/times.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <dirent.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>

#define DFLT_SHELL	"/bin/sh"
#define FILE_SEP_CHAR	'/'

#define closesocket(fd)	    close((fd))

#ifdef HAVE_PTHREADS
#include <pthread.h>

pthread_mutex_t Mutex;
pthread_attr_t ThreadAttr;
#endif
#endif


#define MAX_BUF_SIZE	8192	/* Maximum network buffer size (< 2^15) */
#define MAX_LINE_SIZE	1024	/* Maximum file line size */
#define MAX_KEY_BYTES	((BF_ROUNDS + 2)*4) /* Maximum size of Blowfish key */
#define	MIN_KEY_BYTES	5	/* Minimum key length */
#define MAX_LISTEN	5	/* Depth of listen queue */
#define MAX_INCLUDE	5	/* Maximum depth of include files */

#define HASH_STR_SIZE	41	/* Size of SHA hash string including null */
#define TIMESTAMP_SIZE	20	/* Size of YYYY-dd-mm-HH:MM:SS timestamp */
#define CHALLENGE_SIZE	4	/* Size of challenge data */
#define THE_ANSWER	42	/* To Life, the Universe and Everything */

#define THREAD_STACK_SIZE   65536

#define	CMP_OVERHEAD	800	/* Maximum overhead on 8k message */
#define	CMP_MINIMUM	16	/* Minimum message size to attempt compression */

/*
** Information about the compression algorithm and level is encoded in
** a single unsigned short value. The high 8 bits are the algorithm and
** the low eight bits the level. Note that the values used ensure that
** taken as a 16-bit quantity all bzip2 values are greater than all
** zlib values. This fact is used so that, in effect, bzip2 compression
** is considered "stronger" than zlib.
*/

#define CMPTYPE_ZLIB	0x0
#define CMPTYPE_BZIP2	0x1
#define GET_CMPTYPE(z)	    (((z) >> 8) & 0xff)
#define SET_CMPTYPE(z, t)   ((z) | ((t) << 8))
#define GET_CMPLEVEL(z)	    ((z) & 0xff)
#define SET_CMPLEVEL(z, l)  ((z) | ((l) & 0xff))

/*
** Each message that Zebedee transmits is preceded by an unsigned short
** value (in big-endian format). The top two bits flag whether the message
** is encrypted and compressed. The lower 14 bits define the payload size
** (which must be no greater than MAX_BUF_SIZE -- which is 8192 bytes in
** the current version of the protocol).
*/

#define FLAG_COMPRESSED	0x1
#define FLAG_ENCRYPTED	0x2

#define GET_FLAGS(x)	(((x) >> 14) & 0x3)
#define SET_FLAGS(x, f)	((x) | ((f) << 14))

#define GET_SIZE(x)	((x) & 0x3fff)

#define	INIT_IVEC	"Time4Bed"  /* ... said Zebedee. Boing! */
typedef struct
{
    BF_KEY key;
    unsigned char iVec[8];
    int pos;
    unsigned char cryptBuf[MAX_BUF_SIZE];
}
BFState_t;

typedef struct PortList_s
{
    unsigned short lo;
    unsigned short hi;
    struct PortList_s *next;
}
PortList_t;

#define DFLT_GENERATOR	    "2"
#define DFLT_MODULUS	    \
"f488fd584e49dbcd20b49de49107366b336c380d451d0f7c88b31c7c5b2d8ef6" \
"f3c923c043f0a55b188d8ebb558cb85d38d334fd7c175743a31d186cde33212c" \
"b52aff3ce1b1294018118d7c84a70a72d686c40319c807297aca950cd9969fab" \
"d00a509b0246d3083d66a45d419f9c7cbd894b221926baaba25ec355e92f78c7"
#define DFLT_CMP_LEVEL	    SET_CMPLEVEL(CMPTYPE_ZLIB, 6)
#define DFLT_KEY_BITS	    128
#define DFLT_PORT	    0x2EBD

#define PROTOCOL_VERSION    0x0100  /* The one and only version so far */

/*****************\
**		 **
**  Global Data  **
**		 **
\*****************/

/*
** Note: Although this data is global it is not protected by mutex locks
** because once set in the start-up phases of the program it is read-only
** by the rest of the routines.
*/

FILE *LogFileP = NULL;		/* File handle for log file (NULL => stderr) */
unsigned short LogLevel = 1;	/* Message verbosity level */
char *Program = "zebedee";	/* Program name (argv[0]) */
char *Generator = "";		/* DH generator hex string ("" => default) */
char *Modulus = "";		/* DH modulus hex string ("" => default) */ 
char *PrivateKey = NULL;	/* Private key hex string */
unsigned short KeyLengthBits = DFLT_KEY_BITS;	/* Key length in bits */
unsigned short CompressInfo = DFLT_CMP_LEVEL;	/* Compression type and level */
int IsDetached = 1;		/* Flag true if program should run detached */
int IsServer = 0;		/* Flag true if program is a server */
int Debug = 0;			/* Debug mode -- single threaded server */
char *CommandString = NULL;	/* Command string to execute (client) */
unsigned short ServerPort = DFLT_PORT;	/* Port on which server listens */
unsigned short LocalPort = 0;	/* Port on which client listens */
unsigned short RemotePort = 23;	/* Remote port to which to tunnel */
char *RemoteHost = NULL;	/* Name of remote host */
char *IdentityFile = NULL;	/* Name of identity file to check, if any */
PortList_t *AllowedPorts = NULL;    /* List of allowed redirected ports */
char *KeyGenCmd = NULL;		/* Key generator command string */
int TimestampLog = 0;		/* Should messages have timestamps? */
int MultiUse = 0;		/* Client handles multiple connections? */

extern char *optarg;		/* From getopt */
extern int optind;		/* From getopt */


/*************************\
**			 **
**  Function Prototypes  **
**			 **
\*************************/

void threadInit(void);
void threadLock(void);
void threadUnlock(void);
unsigned long threadPid(void);
unsigned long threadTid(void);

void timestamp(char *timeBuf, int local);
void message(unsigned short level, int err, char *fmt, ...);

int readData(int fd, unsigned char *buffer, unsigned short size);
int readUShort(int fd, unsigned short *resultP);
int writeData(int fd, unsigned char *buffer, unsigned short size);
int writeUShort(int fd, unsigned short value);

int readMessage(int fd, unsigned char *buffer, unsigned short *sizeP,
		unsigned short cmpInfo, BFState_t *bfStateP);
int writeMessage(int fd, unsigned char *buffer, unsigned short size,
		 unsigned short cmpInfo, BFState_t *bfStateP);
int requestResponse(int fd, unsigned short request, unsigned short *responseP);

int makeConnection(const char *host, const unsigned short port);
int makeListener(unsigned short *portP, int localOnly);

BFState_t *setupBlowfish(char *keyStr, unsigned short keyBits);
char *generateKey(void);
char *runKeyGenCommand(void);
unsigned short hexStrToBits(char *hexStr, unsigned short bits, unsigned char *bitVec);
char *diffieHellman(char *genStr, char *modStr, char *expStr);
void makeChallenge(unsigned char *challenge);
void challengeAnswer(unsigned char *challenge);

int spawnCommand(unsigned short port, char *cmdFormat);
int filterLoop(int localFd, int remoteFd, BFState_t *sendBFP,
	       BFState_t *recvBFP, unsigned short cmpInfo);


void hashStrings(char *hashBuf, ...);
void hashFile(char *hashBuf, char *fileName);
int checkIdentity(char *idFile, char *generator, char *modulus, char *key);
char *generateIdentity(char *generator, char *modulus, char *exponent);

void spawnHandler(void (*handler)(int), int listenFd, int clientFd, int inLine);
void clientListener(unsigned short port);
void client(int clientFd);
void makeDetached(void);
void serverListener(unsigned short port);
int allowRedirect(unsigned short port);
void server(int clientFd);

unsigned short scanPortRange(const char *str, unsigned short *loP,
			     unsigned short *hiP);
void setBoolean(char *value, int *resultP);
void setUShort(char *value, unsigned short *resultP);
void setPort(char *value, unsigned short *resultP);
void setAllowedPorts(char *value);
void setString(char *value, char **resultP);
void setLogFile(char *newFile);
void setCmpInfo(char *value, unsigned short *resultP);
void readConfigFile(const char *fileName, int level);

char *cleanHexString(char *str);

void usage(void);

void sigpipeHandler(int sig);

/*************************************\
**				     **
**  Thread Synchronisation Routines  **
**				     **
\*************************************/

/*
** threadInit
**
** Initialise global mutex and thread attributes. Must be called before
** any other thread routines.
*/

void
threadInit(void)
{
#if defined(WIN32)
    InitializeCriticalSection(&Mutex);
#elif defined(HAVE_PTHREADS)
    pthread_mutex_init(&Mutex, NULL);
    pthread_attr_init(&ThreadAttr);
    pthread_attr_setstacksize(&ThreadAttr, THREAD_STACK_SIZE);
    pthread_attr_setdetachstate(&ThreadAttr, PTHREAD_CREATE_DETACHED);
#endif
}

/*
** threadLock
**
** Lock the global mutex
*/

void
threadLock(void)
{
#if defined(WIN32)
    EnterCriticalSection(&Mutex);
#elif defined(HAVE_PTHREADS)
    pthread_mutex_lock(&Mutex);
#endif
}

/*
** threadUnlock
**
** Unlock the global mutex
*/

void
threadUnlock(void)
{
#if defined(WIN32)
    LeaveCriticalSection(&Mutex);
#elif defined(HAVE_PTHREADS)
    pthread_mutex_unlock(&Mutex);
#endif
}

/*
** threadPid
**
** Return the current process ID
*/

unsigned long
threadPid(void)
{
#ifdef WIN32
    return (unsigned long)GetCurrentProcessId();
#else
    return (unsigned long)getpid();
#endif
}

/*
** threadTid
**
** Return the current thread ID
*/

unsigned long
threadTid(void)
{
#ifdef WIN32
    return (unsigned long)GetCurrentThreadId();
#elif defined(HAVE_PTHREADS)
    return (unsigned long)pthread_self();
#else
    return 0;
#endif
}

/*********************\
**		     **
**  Message Logging  **
**		     **
\*********************/

void
timestamp(char *timeBuf, int local)
{
    time_t now;
    struct tm *tmPtr;

    threadLock();
    time(&now);
    if (local)
    {
	tmPtr = localtime(&now);
    }
    else
    {
	tmPtr = gmtime(&now);
    }
    strftime(timeBuf, TIMESTAMP_SIZE, "%Y-%m-%d-%H:%M:%S", tmPtr);
    threadUnlock();
}

#ifdef WIN32
void
logAppEvent(unsigned short level, char *msg)
{
    HANDLE  eventHandle;
    char    *strings[2];

    eventHandle = RegisterEventSource(NULL, Program);

    strings[0] = msg;
    strings[1] = NULL;

    if (eventHandle != NULL)
    {
	ReportEvent(eventHandle,	    /* Handle of event source */
		    (level ? EVENTLOG_INFORMATION_TYPE :
		     EVENTLOG_ERROR_TYPE),  /* Event type */
		    (WORD)level,	    /* Event category */
		    0,			    /* Event ID */
		    NULL,		    /* User SID */
		    1,			    /* Number of message strings */
		    0,			    /* Bytes of binary data */
		    (const char **)strings, /* Array of message strings */
		    NULL);		    /* No binary data */
	DeregisterEventSource(eventHandle);
    }
}
#endif

/*
** message
**
** Output a message to the current log file if the message verbosity is
** greater than or equal to the specified level. Messages at level 0
** can not be suppressed and are all error messages.
**
** If errno is non-zero then append the matching error text.
*/

void
message(unsigned short level, int err, char *fmt, ...)
{
    FILE *fp = LogFileP;
    va_list args;
    char timeBuf[TIMESTAMP_SIZE];
    char *timePtr = NULL;
    char msgBuf[MAX_LINE_SIZE];


    if (level > LogLevel) return;

    va_start(args, fmt);

    /* Several thread-unsafe routines are called beyond here */

    
    if (fp == NULL)
    {
	fp = stderr;
    }

    if (TimestampLog)
    {
	timestamp(timeBuf, 1);
	timePtr = timeBuf;
    }

    /* Ensure we don't get overlapping messages */

    threadLock();

    /*
    ** The message format is the program name followed by the (low five
    ** digits of) the PID and thread ID then an optional timestamp followed
    ** by an amount of indentation determined by the level. This is
    ** then followed by the supplied message text and arguments and
    ** finally the error message text (if any) associated with the supplied
    ** error number!
    */

    sprintf(msgBuf, "%s(%lu/%lu): %s%s%.*s%s",
	    Program, (threadPid() % 100000), (threadTid() % 100000),
	    (timePtr ? timePtr : ""), (timePtr ? ": " : ""),
	    level, "          ", (level ? "" : "ERROR: "));

    vsprintf(msgBuf + strlen(msgBuf), fmt, args);

    va_end(args);

    if (err)
    {
	sprintf(msgBuf + strlen(msgBuf), ": (%s)", strerror(err));
    }

    fprintf(fp, "%s\n", msgBuf);

    fflush(fp);

    threadUnlock();

#ifdef WIN32
    /* Messages at levels <= 3 also go to the system application event log */

    if (level <= 3)
    {
	logAppEvent(level, msgBuf);
    }
#endif
}

/*******************************\
**			       **
**  Network Data Transmission  **
**			       **
\*******************************/

/*
** readData
**
** Read and reassemble a potentially fragmented message from the network.
*/

int
readData(int fd, unsigned char *buffer, unsigned short size)
{
    int num = 0;
    char *bufP = NULL;
    unsigned short total = 0;

    bufP = (char *)buffer;
    do
    {
	message(5, 0, "readData: receiving %d of %d", (size - total), size);
	if ((num = recv(fd, (bufP + total), (size - total), 0)) <= 0)
	{
	    message(5, errno, "readData: EOF or error");
	    /* Premature EOF or error */
	    return num;
	}
	message(5, 0, "readData: read %d byte(s)", num);
	total += (unsigned short)num;
    }
    while (total < size);

    return total;
}

/* 
** readUShort
**
** Read an unsigned short value from the network.
**
** The value is transmitted in big-endian format. The routine returns the
** number of bytes read (or 0 on EOF, -1 on error) and the value itself
** via valueP.
*/

int
readUShort(int fd, unsigned short *resultP)
{
    int num = 0;
    unsigned char buffer[2];

    if ((num = readData(fd, buffer, 2)) != 2)
    {
	return num;
    }

    *resultP = ((unsigned short)buffer[0] << 8) + (unsigned short)buffer[1];
    message(4, 0, "readUShort: read %hu", *resultP);

    return num;
}

/*
** writeData
**
*/

int
writeData(int fd, unsigned char *buffer, unsigned short size)
{
    int num = 0;
    char *bufP = NULL;
    unsigned short total = 0;

    bufP = (char *)buffer;
    do
    {
	message(5, 0, "writeData: sending %d of %d", (size - total), size);
	if ((num = send(fd, (bufP + total), (size - total), 0)) <= 0)
	{
	    /* Premature EOF or error */
	    message(5, errno, "writeData: EOF or error");
	    return num;
	}
	total += (unsigned short)num;
	message(5, 0, "writeData: sent %d byte(s)", num);
    }
    while (total < size);

    return total;
}

/*
** writeUShort
**
** Write an unsigned short value to the network in big-endian format
*/

int
writeUShort(int fd, unsigned short value)
{
    unsigned char buf[2];

    message(4, 0, "writeUShort: writing %hu", value);

    buf[0] = (unsigned char)((value >> 8) & 0xff);
    buf[1] = (unsigned char)(value & 0xff);

    return writeData(fd, buf, 2);
}

/*
** readMessage
**
** Read a message from the network into the supplied buffer, uncompressing
** and decrypting as necessary.
**
** The size of the buffer on input is given in sizeP. The agreed compression
** type (and level) is in cmpInfo. The current encryption context is given
** via bfStateP.
**
** The size of the expanded, unencrypted message is returned as the value
** of the function and also via sizeP. If there is an error then -1 is
** returned.
*/

int
readMessage(int fd, unsigned char *buffer, unsigned short *sizeP,
	    unsigned short cmpInfo, BFState_t *bfStateP)
{
    unsigned short hdr;
    unsigned short size;
    unsigned short flags;
    unsigned char inBuffer[MAX_BUF_SIZE];
    int num = 0;
    unsigned long uncmpSize = MAX_BUF_SIZE;


    /* Read the header */

    if ((num = readUShort(fd, &hdr)) != 2) return num;

    /* Extract the flags and message size */

    flags = GET_FLAGS(hdr);
    size = GET_SIZE(hdr);

    /* Reject invalid messages */

    if (size > MAX_BUF_SIZE)
    {
	message(0, 0, "incoming message size too big (%hu > %hu)", size, MAX_BUF_SIZE);
	return -1;
    }

    assert(size <= MAX_BUF_SIZE && size <= *sizeP);
    *sizeP = size;

    message(4, 0, "readMessage: message size = %hu, %s, %s", size,
	    ((flags & FLAG_ENCRYPTED) ? "encrypted" : "unencrypted"),
	    ((flags & FLAG_COMPRESSED) ? "compressed" : "uncompressed"));

    /* Read the remaining message data */

    if ((num = readData(fd, inBuffer, size)) != (int)size) return num;

    /* Decrypt if necessary */

    if (flags & FLAG_ENCRYPTED)
    {
	assert(bfStateP != NULL);

	BF_cfb64_encrypt(inBuffer, bfStateP->cryptBuf, size,
			 &(bfStateP->key), bfStateP->iVec, &(bfStateP->pos),
			 BF_DECRYPT);
	memcpy(inBuffer, bfStateP->cryptBuf, size);
    }

    /* Decompress if necessary */

    if (flags & FLAG_COMPRESSED)
    {
	switch (GET_CMPTYPE(cmpInfo))
	{
	case CMPTYPE_ZLIB:
	    if ((num = uncompress(buffer, &uncmpSize,
				  (Byte *)inBuffer, size)) != Z_OK)
	    {
		message(0, errno, "uncompressing message data (zlib status = %d)", num);
		errno = 0;
		return -1;
	    }
	    break;

	case CMPTYPE_BZIP2:
#ifndef DONT_HAVE_BZIP2
	    if ((num = bzBuffToBuffDecompress((char *)buffer,
					      (unsigned int *)&uncmpSize,
					      (char *)inBuffer,
					      (unsigned int)size,
					      0, 0)) != BZ_OK)
	    {
		message(0, errno, "uncompressing message data (bzip2 status = %d)", num);
		errno = 0;
		return -1;
	    }
	    break;
#else
	    message(0, 0, "received unsupported bzip2 compressed message -- should never happen!");
	    return -1;
	    break;
#endif
	default:
	    message(0, 0, "invalid compression info in readMessage (%#hx)", cmpInfo);
	    return -1;
	    break;
	}

	*sizeP = size = (unsigned short)uncmpSize;
	message(4, 0, "readMessage: uncompressed size = %hu", size);
    }
    else
    {
	memcpy(buffer, inBuffer, size);
    }

    return (int)size;
}

/*
** writeMessage
**
** Write a message to the network containing the data from buffer, compressing
** and encrypting as necessary.
**
** The size of the input buffer is given in size. The desired compression
** type and level (which may be zero) is specified in cmpInfo and the current
** encryption context is given in bfStateP. If the message is not to
** be encrypted then bfCryptP should be NULL.
**
** The size of the original expanded, unencrypted message is returned as the
** value of the function on success or the status from writeData() on error.
**
** The message is sent as an unsigned short header (in the format written
** by writeUShort) followed by the data itself. The header value consists
** of the length or'ed with flags indicating if the message is compressed
** and encrypted.
*/

int
writeMessage(int fd, unsigned char *buffer, unsigned short size,
	     unsigned short cmpInfo, BFState_t *bfStateP)
{
    unsigned short origSize = size;
    unsigned short hdr;
    unsigned char outBuffer[MAX_BUF_SIZE + CMP_OVERHEAD + 2];
    int num = 0;
    unsigned long cmpSize = MAX_BUF_SIZE + CMP_OVERHEAD;
    unsigned short flags = 0;


    /* Attempt compression if the message size warrants it */

    if (cmpInfo && size > CMP_MINIMUM)
    {
	switch (GET_CMPTYPE(cmpInfo))
	{
	case CMPTYPE_ZLIB:
	    if ((num = compress2(outBuffer + 2, &cmpSize,
				 (const Byte *)buffer, size,
				 GET_CMPLEVEL(cmpInfo))) != Z_OK)
	    {
		message(0, errno, "compressing data (zlib status = %d)", num);
		cmpSize = size;
	    }
	    break;

	case CMPTYPE_BZIP2:
#ifndef DONT_HAVE_BZIP2
	    if ((num = bzBuffToBuffCompress((char *)(outBuffer + 2),
					    (unsigned int *)&cmpSize,
					    (char *)buffer,
					    (unsigned int)size,
					    (int)GET_CMPLEVEL(cmpInfo),
					    0, 0)) != BZ_OK)
	    {
		message(0, errno, "compressing data (bzip2 status = %d)", num);
	    }
	    break;
#else
	    message(0, 0, "request to use unsupported bzip2 compression!");
	    cmpSize = (unsigned long)size;
	    break;
#endif

	default:
	    cmpSize = (unsigned long)size;
	    break;
	}

	/* Only use compressed message if it is shorter */

	if (cmpSize < (unsigned long)size)
	{
	    message(4, 0, "writeMessage: message compressed from %hu to %lu bytes", size, cmpSize);
	    buffer = outBuffer + 2;
	    size = (unsigned short)cmpSize;
	    flags |= FLAG_COMPRESSED;
	}
    }

    /* Encrypt if required */

    if (bfStateP)
    {
	BF_cfb64_encrypt(buffer, bfStateP->cryptBuf, size,
			 &(bfStateP->key), bfStateP->iVec, &(bfStateP->pos),
			 BF_ENCRYPT);
	memcpy(outBuffer + 2, bfStateP->cryptBuf, size);
	flags |= FLAG_ENCRYPTED;
    }
    else
    {
	memmove(outBuffer + 2, buffer, size);
    }

    /* Insert the header */

    hdr = SET_FLAGS(size, flags);

    outBuffer[0] = (unsigned char)((hdr >> 8) & 0xff);
    outBuffer[1] = (unsigned char)(hdr & 0xff);

    /* Write the message data */

    message(4, 0, "writeMessage: message size = %hu, %s, %s", size,
	    ((flags & FLAG_ENCRYPTED) ? "encrypted" : "unencrypted"),
	    ((flags & FLAG_COMPRESSED) ? "compressed" : "uncompressed"));

    if ((num = writeData(fd, outBuffer, size + 2)) != (int)(size + 2)) return num;

    return origSize;
}

/*
** requestResponse
**
** This is a helper routine that sends the unsigned short "request" value
** and awaits the response, storing it in "responseP". It returns 1 if
** successful or 0 otherwise.
*/

int
requestResponse(int fd, unsigned short request, unsigned short *responseP)
{
    /* Write request */

    if (writeUShort(fd, request) != 2)
    {
	return 0;
    }

    /* Read response */

    if (readUShort(fd, responseP) != 2)
    {
	return 0;
    }

    return 1;
}

/*
** makeConnection
**
** Set up a socket connection to the specified host and port. The host
** name can either be a DNS name or a string IP address.
*/

int
makeConnection(const char *host, const unsigned short port)
{
    struct hostent *entry = NULL;
    int sfd = -1;
    struct sockaddr_in addr;


    /* Sanity check */

    assert(host != NULL && port != 0);

    /* Translate hostname from DNS or IP-address form */

    if ((entry = gethostbyname(host)) == NULL)
    {
	if ((addr.sin_addr.s_addr = htonl(inet_addr(host))) == 0xffffffff)
	{
	    message(0, 0, "can't resolve host or address '%s'", host);
	    errno = 0;
	    return -1;
	}
    }
    else
    {
	memcpy(&addr.sin_addr, entry->h_addr, entry->h_length);
    }
    addr.sin_family = AF_INET;
    addr.sin_port = ntohs(port);

    if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	message(0, errno, "socket creation failed");
	errno = 0;
	return -1;
    }

    if (connect(sfd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    {
	closesocket(sfd);
	return -1;
    }

    return sfd;
}

/*
** makeListener
**
** Set up a listener socket on the port supplied via portP. If localOnly
** is true then the socket will only be set up for listening to connections
** from localhost.
**
** If the requested port is 0 then a new port will be allocated and
** returned in portP. If the requested port is non-zero then an attempt
** will be made to re-use the address if possible.
**
** The routine returns the socket ID on success or -1 on error.
*/

int
makeListener(unsigned short *portP, int localOnly)
{
    int sfd = -1;
    struct sockaddr_in addr;
    int addrLen = sizeof(addr);
    int trueVal = 1;


    /* Create the socket */

    if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	message(0, errno, "can't create listener socket");
	goto failure;
    }

    /* If we requested a specific port then reuse the address if possible */

    if (*portP)
    {
	(void)setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (char *)&trueVal, sizeof(trueVal));
    }

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = localOnly ? htonl(INADDR_LOOPBACK) : htonl(INADDR_ANY);
    addr.sin_port = htons(*portP);

    if (bind(sfd, (struct sockaddr *)&addr, addrLen) < 0)
    {
	message(0, errno, "listener bind failed");
	goto failure;
    }

    if (listen(sfd, MAX_LISTEN) < 0)
    {
	message(0, errno, "listen failed");
	goto failure;
    }

    /* Retrieve the port actually being used to return via portP */

    if (getsockname(sfd, (struct sockaddr *)&addr, &addrLen))
    {
	message(0, errno, "can't get local port number");
	goto failure;
    }
    *portP = ntohs(addr.sin_port);

    return sfd;

failure:
    if (sfd != -1)
    {
	(void)closesocket(sfd);
    }
    errno = -1;
    return -1;
}

/*********************************\
**				 **
**  Encryption-related Routines  **
**				 **
\*********************************/

/*
** setupBlowfish
**
** Create and initialise a BFState_t structure to hold the encryption
** context for one communication stream (A -> B or B -> A but not both).
**
** keyStr is the key data which is in the form of a string of hexadecimal
** digits. The actual key used is the high-order keyBits bits of this
** number.
**
** The routine returns a pointer to the newly-allocated structure on success
** or NULL on error. As a special case, if the key length is zero then no
** encryption is needed and we also return NULL.
*/

BFState_t *
setupBlowfish(char *keyStr, unsigned short keyBits)
{
    BFState_t *bf = NULL;
    unsigned char keyData[MAX_KEY_BYTES];
    int keyBytes;


    /* Special case -- no encryption requeste */

    if (keyBits == 0) return NULL;

    /* Now allocate the necessary space */

    if ((bf = (BFState_t *)malloc(sizeof(BFState_t))) == NULL)
    {
	message(0, errno, "out of memory allocating Blowfish state data");
	errno = 0;
	return NULL;
    }

    keyBytes = hexStrToBits(keyStr, keyBits, keyData);

    memset(bf, 0, sizeof(BFState_t));
    BF_set_key(&(bf->key), keyBytes, keyData);
    memcpy(bf->iVec, INIT_IVEC, 8);
    bf->pos = 0;

    return bf;
}

/*
** generateKey
**
** Generate the exponent (private key) for the Diffie-Hellman key
** exchange.
**
** Good key generation is crucial for the security of the encryption
** mechanism. Ideally we would generate keys based on truly random
** sources of data -- like radioactive decay. Unfortunately this is
** not tremendously practical so we have to do the best we can to
** generate keys that will be hard to predict. How well we can achieve
** this depends somewhat on the operating system on which the program
** is running. See the inline comments below.
**
** In the comments we will try to set some bounds on the number of
** "bits of uncertainty" (BOU) -- that is how many bits' worth of
** imprecision there is in determining various quantities for an attacker.
** The "Min BOU" is where an attacker has access to the system and can
** examine system performance counters etc. The "Max BOU" is where there
** is no such access and all the attacker has access to is the data "on
** the wire". Please note that these are estimates, NOT guarantees!
**
** Basically, if you are happy that an attacker can not see the state of
** your system when the key is generated (either through a direct login
** or via remote administrative interfaces) then the algorithms below
** are PROBABLY sufficient but in other cases you will need a
** different approach.
**
** If you are not happy with these key generation mechanisms then
** you can call out to an external program. This needs to generate a
** single string of hexadecimal digits (at least MIN_KEY_BYTES long)
** which will be used as the key.
*/

char *
generateKey(void)
{
    SHA_INFO sha;
    time_t now = time(NULL);
    unsigned long pid = (unsigned long)getpid();
    char *result = NULL;


    /* If a private key has been supplied copy it and return it */

    if (PrivateKey)
    {
	if ((result = (char *)malloc(strlen(PrivateKey) + 1)) == NULL)
	{
	    return NULL;
	}
	strcpy(result, PrivateKey);
	return result;
    }

    /*
    ** If a key generator command was specified then use this to generate
    ** the key rather than doing it directly here. If the generator fails,
    ** however, then we will fall back to the inline method.
    */

    if ((result = runKeyGenCommand()) != NULL)
    {
	return result;
    }

    /*
    ** We use SHA to "stir" whatever bits of "entropy" we can acquire.
    ** This distributes the input very well over 160 bits.
    */

    sha_init(&sha);

    /*
    ** In all cases add the current time and process ID. The time can be
    ** guessed by an attacker to within a second. With physical access
    ** to the machine the PID can be obtained with certainty. Even without
    ** physical access on UNIX the uncertainty in the PID is less than 16
    ** bits. On Win32 it seems even less.
    **
    ** Min BOU: 1
    ** Max BOU: 8
    */

    sha_update(&sha, (SHA_BYTE *)&now, sizeof(now));
    sha_update(&sha, (SHA_BYTE *)&pid, sizeof(pid));

#if defined(WIN32)
    {
	LARGE_INTEGER perf;
	DWORD ticks = timeGetTime();
	FILETIME created, exited, kernel, user;


	/*
	** The "ticks" value is the elapsed time in milliseconds since
	** Windows was started.
	**
	** Assume that the time at which the key is generated can be
	** predicted within 0.01 sec and that the clock resolution is 10ms.
	**
	** Min BOU: 2
	** Max BOU: 32
	*/

	sha_update(&sha, (SHA_BYTE *)&ticks, sizeof(ticks));

	/*
	** QueryPerformanceCounter gives a very high resolution 64-bit
	** time result. Unfortunately, if there is no hardware support
	** for a high-resolution timer it can return zero. On hardware
	** I have available the resolution is over 1 million counts per
	** second.
	**
	** Assume the same accuracy of guessing as above.
	**
	** Min BOU: 12 (but not independent from the previous quantity).
	** Max BOU: 64
	*/

	(void)QueryPerformanceCounter(&perf);

	sha_update(&sha, (SHA_BYTE *)&perf, sizeof(perf));

	/*
	** The following quantities are 64 bit times in 100nsec
	** intervals since Jan 1, 1601. They are available to be
	** read by other suitably privileged processes. I'm not sure
	** of the resolution and only the kernel and user times
	** have any degree of unpredictability from "outside" so
	** we will make conservative estimates.
	**
	** Min BOU: 4
	** Max BOU: 32
	*/

	GetProcessTimes(GetCurrentProcess(),
			&created, &exited, &kernel, &user);

	sha_update(&sha, (SHA_BYTE *)&created, sizeof(created));
	sha_update(&sha, (SHA_BYTE *)&exited, sizeof(exited));
	sha_update(&sha, (SHA_BYTE *)&kernel, sizeof(kernel));
	sha_update(&sha, (SHA_BYTE *)&user, sizeof(user));

	/*
	** Total estimates for Win32
	**
	** Min BOU: 17, Max BOU: 136
	*/
    }
#else	/* !WIN32 */
    {
	clock_t ticks;
	struct tms tms;

	/*
	** On all UNIX systems we get the process time stats. These are
	** 32-bit quantities relative to system boot.
	**
	** Min BOU: 2
	** Max BOU: 40
	*/

	ticks = times(&tms);

	sha_update(&sha, (SHA_BYTE *)&ticks, sizeof(ticks));
	sha_update(&sha, (SHA_BYTE *)&tms, sizeof(tms));
    }
#if defined(HAVE_DEV_RANDOM)
    {
	/*
	** Now we're talking! /dev/random uses internal kernel counters
	** and state not accessible to normal users. We will read 10 chunks
	** of 8 bytes which should give us more than enough to justify
	** claiming that we have a full 160 bits of uncertainty coming
	** out of the final hash.
	**
	** BOU: 160+
	*/

	int fd = open("/dev/random", O_RDONLY);
	char buffer[8];
	int i;

	for (i = 0; i < 10; i++)
	{
	    read(fd, buffer, 8);
	    sha_update(&sha, (SHA_BYTE *)buffer, 8);
	}
	close(fd);
    }
#elif defined(HAVE_PROC_FS)
    {
	/*
	** If we haven't got /dev/random but do have /proc then we can
	** probably do quite well on an active system. We stat every
	** process and hash that data in. On a very stable system this
	** could, however, be fairly predictable to an attacker with
	** access to the system.
	**
	** Min BOU: 4
	** Max BOU: 160+
	*/

	struct stat sbuf;
	struct dirent *entryP;
	struct dirent entryBuf;
	DIR *dir = opendir("/proc");
	char name[MAX_LINE_SIZE];


	if (dir == NULL)
	{
	    message(0, errno, "can't open /proc directory");
	}
	else
	{
	    while ((entryP = readdir(dir)) != NULL)
	    {
		sprintf(name, "/proc/%s", entryP->d_name);
		stat(name, &sbuf);
		sha_update(&sha, (SHA_BYTE *)entryP, sizeof(struct dirent));
		sha_update(&sha, (SHA_BYTE *)&sbuf, sizeof(sbuf));
	    }
	    closedir(dir);
	}
    }
#endif	/* HAVE_PROC_FS */
#endif	/* !WIN32 */

    /* Convert the digest to a string and return */

    sha_final(&sha);

    if ((result = (char *)malloc(HASH_STR_SIZE)) == NULL)
    {
	return NULL;
    }

    sprintf(result, "%08lx%08lx%08lx%08lx%08lx", sha.digest[0], sha.digest[1],
	    sha.digest[2], sha.digest[3], sha.digest[4]);

    return result;
}

/*
** runKeyGenCommand
**
** This runs the external key generation command, if any. It reads a single
** line from the command's standard output and extracts a hex string from
** it. The string must be at least MIN_KEY_BYTES * 2 bytes long.
**
** If the routine succeeds it malloc's space for the string and returns
** it otherwise it returns NULL.
*/

char *
runKeyGenCommand(void)
{
    FILE *fp;
    char buf[MAX_LINE_SIZE];
    char *result = NULL;


    if (KeyGenCmd == NULL)
    {
	message(3, 0, "no key generation command specified");
	return NULL;
    }

    if ((fp = popen(KeyGenCmd, "r")) == NULL)
    {
	message(0, errno, "failed to spawn key generation command '%s'", KeyGenCmd);
	return NULL;
    }

    if (fgets(buf, MAX_LINE_SIZE, fp) != NULL)
    {
	if ((result = (char *)malloc(strlen(buf) + 1)) != NULL)
	{
	    if (sscanf(buf, "%[0-9a-fA-F]", result) != 1 ||
		strlen(buf) < (MIN_KEY_BYTES * 2))
	    {
		free(result);
		result = NULL;
	    }
	}
    }

    fclose(fp);

    message(3, 0, "key generation result ", (result ? "not null" : "NULL"));
    return result;
}

/*
** hexStrToBits
**
** This converts the first "bits" bits of hex string "hexStr" to a
** bit vector in "bitVec", which must be at least MAX_KEY_BYTES bytes
** long. It returns the number of bytes of bitVec that were set.
**
** Note that "bits" is effectively rounded up to the nearest multiple
** of 4.
*/

unsigned short
hexStrToBits(char *hexStr, unsigned short bits, unsigned char *bitVec)
{
    int i;
    unsigned char byte;
    int len = strlen(hexStr);


    memset(bitVec, 0, MAX_KEY_BYTES);

    /* Determine number of nybbles required */

    if ((int)((bits + 3) / 4U) < len)
    {
	len = (int)((bits + 3) / 4U);
    }

    /* Truncate the number of nybbles to fit in the buffer, if necessary */

    if (len > (MAX_KEY_BYTES * 2))
    {
	len = MAX_KEY_BYTES * 2;
    }

    /* Now process the string a nybble at a time */

    for (i = 0; i < len; i += 2)
    {
	byte = '\0';

	/* High nybble */

	if (hexStr[i] >= '0' && hexStr[i] <= '9')
	{
	    byte = (hexStr[i] - '0') << 4;
	}
	else if (toupper(hexStr[i]) >= 'A' && toupper(hexStr[i]) <= 'F')
	{
	    byte = (toupper(hexStr[i]) - 'A' + 0xA) << 4;
	}

	/* Low nybble -- if any, otherwise left as zero */

	if (i + 1 < len)
	{
	    if (hexStr[i + 1] >= '0' && hexStr[i + 1] <= '9')
	    {
		byte |= (hexStr[i + 1] - '0');
	    }
	    else if (toupper(hexStr[i + 1]) >= 'A' && toupper(hexStr[i + 1]) <= 'F')
	    {
		byte |= (toupper(hexStr[i + 1]) - 'A' + 0xA);
	    }
	}

	bitVec[i / 2] = byte;
    }

    /* Return number of bytes */

    return (unsigned short)(i / 2);
}

/*
** diffieHellman
**
** Perform the core Diffie-Hellman calculation which is
**
**	(generator ** exponent) mod modulus
**
** This operates on hex strings and returns a newly-allocated hex string
** as its answer. If genStr or modStr are NULL or empty strings then
** the default, built-in values will be used.
*/

char *
diffieHellman(char *genStr, char *modStr, char *expStr)
{
    mpz_t gen, mod, exp, key;
    char *keyStr = NULL;


    if (genStr == NULL || *genStr == '\0') genStr = DFLT_GENERATOR;
    if (modStr == NULL || *modStr == '\0') modStr = DFLT_MODULUS;

    mpz_init_set_str(gen, genStr, 16);
    mpz_init_set_str(mod, modStr, 16);
    mpz_init_set_str(exp, expStr, 16);
    mpz_init(key);

    mpz_powm(key, gen, exp, mod);

    keyStr = mpz_get_str(NULL, 16, key);

    mpz_clear(gen);
    mpz_clear(exp);
    mpz_clear(mod);
    mpz_clear(key);

    return keyStr;
}

/*
** makeChallenge
**
** Create a challenge value and write it into "challenge", which
** must be CHALLENGE_SIZE bytes long. We will just use clock ticks.
*/

void
makeChallenge(unsigned char *challenge)
{
#ifdef WIN32
    DWORD ticks = timeGetTime();
    memcpy(challenge, &ticks, CHALLENGE_SIZE);
#else
    struct tms tms;
    clock_t ticks = times(&tms);
    memcpy(challenge, &ticks, CHALLENGE_SIZE);
#endif
}

/*
** challengeAnswer
**
** Calculate the answer to the challenge. XOR each byte with THE_ANSWER.
*/

void
challengeAnswer(unsigned char *challenge)
{
    int i;

    for (i = 0; i < CHALLENGE_SIZE; i++)
    {
	challenge[i] ^= THE_ANSWER;
    }
}

/*****************************\
**			     **
**  General Helper Routines  **
**			     **
\*****************************/

/*
** spawnCommand
**
** Start a sub-process running the specified command using the default
** shell. The command string may contain a single "%d" format character
** which will be replaced with the local port number.
*/

int
spawnCommand(unsigned short port, char *cmdFormat)
{
#ifdef WIN32
    char cmdBuf[MAX_LINE_SIZE];
    STARTUPINFO	suInfo;
    PROCESS_INFORMATION pInfo;

    sprintf(cmdBuf, cmdFormat, (int)port);

    memset(&suInfo, 0, sizeof(suInfo));
    suInfo.cb = sizeof(suInfo);


    if (!CreateProcess(NULL,	/* No executable -- take it from cmdBuf */
		       cmdBuf,	/* Command and arguments */
		       NULL,	/* No security attributes */
		       NULL,	/* No thread attributes */
		       FALSE,	/* Don't inherit handles */
		       0,	/* No special creation flags */
		       NULL,	/* Inherit environment */
		       NULL,	/* Inherit current directory */
		       &suInfo,	/* Start-up info */
		       &pInfo))	/* Process info needed */
    {
	message(0, errno, "failed to spawn '%s'", cmdBuf);
	return 0;
    }
#else
    char *shell = DFLT_SHELL;
    char cmdBuf[MAX_LINE_SIZE];

    sprintf(cmdBuf, cmdFormat, (int)port);

    if (((shell = getenv("SHELL")) == NULL) || *shell == '\0')
    {
	shell = DFLT_SHELL;
    }

    switch (fork())
    {
    case -1:
	message(0, errno, "fork failed");
	return 0;
	break;

    case 0:
	execl(shell, shell, "-c", cmdBuf, NULL);
	message(0, errno, "failed to exec '%s -c \"%s\"'", shell, cmdBuf);
	break;

    default:
	break;
    }
#endif

    return 1;
}

/*
** filterLoop
**
** This is the main processing loop of both client and server. It loops
** reading data from the local system, encrypts and compresses it and
** sends it to the remote system. Conversely it reads data from the remote
** system, decrypts and uncompresses it and writes it to the local system.
**
** On a normal EOF (on either end) it returns 1 otherwise it returns 0.
*/

int
filterLoop(int localFd, int remoteFd,
	   BFState_t *sendBFP, BFState_t *recvBFP,
	   unsigned short cmpInfo)
{
    fd_set testSet;
    int ready = 0;
    int maxTestFd = (localFd > remoteFd ? (localFd + 1) : (remoteFd + 1));
    int num = 0;
    unsigned char buffer[MAX_BUF_SIZE];
    int leave = 0;
    unsigned short size = 0;


    do
    {
	/* Set up file descriptors in mask to test */

	FD_ZERO(&testSet);
	if (localFd >= 0)
	{
	    FD_SET(localFd, &testSet);
	}
	if (remoteFd >= 0)
	{
	    FD_SET(remoteFd, &testSet);
	}

	/* Do a blocking select waiting for any i/o */

	ready = select(maxTestFd, &testSet, 0, 0, 0);

	/* If we get zero then there is nothing left on either fd. */

	if (ready == 0)
	{
	    return 1;
	}

	/* Check for error but ignore interrupted system calls */

	if (ready < 0 && errno != EINTR)
	{
	    message(0, errno, "error in select");
	    return 0;
	}

	/* Is there local data ready? */

	if (FD_ISSET(localFd, &testSet))
	{
	    /* Read a chunk and then write it encrypted and compressed */

	    num = recv(localFd, (char *)buffer, MAX_BUF_SIZE, 0);
	    if (num > 0)
	    {
		if (writeMessage(remoteFd, buffer, (unsigned short)num,
				 cmpInfo, sendBFP) != num)
		{
		    return 0;
		}
	    }
	    else
	    {
		leave++;
	    }
	}

	/* Is there remote data ready? */

	if (FD_ISSET(remoteFd, &testSet))
	{
	    /* Read the encrypted/compressed message and write to local socket */

	    size = MAX_BUF_SIZE;
	    num = readMessage(remoteFd, buffer, &size, cmpInfo, recvBFP);
	    if (num > 0)
	    {
		send(localFd, (char *)buffer, size, 0);
	    }
	    else
	    {
		leave++;
	    }
	}
    }
    while (!leave);

    return 1;
}

/*
** hashStrings
**
** Hash the supplied string arguments (up to a NULL pointer) together and
** write the resulting hash string into hashBuf.
*/

void
hashStrings(char *hashBuf, ...)
{
    SHA_INFO sha;
    va_list ap;
    char *str;

    sha_init(&sha);
    va_start(ap, hashBuf);
    while ((str = va_arg(ap, char *)) != NULL)
    {
	sha_update(&sha, (SHA_BYTE *)str, strlen(str));
    }
    sha_final(&sha);
    sprintf(hashBuf, "%08lx%08lx%08lx%08lx%08lx", sha.digest[0], sha.digest[1],
	    sha.digest[2], sha.digest[3], sha.digest[4]);
}

/*
** hashFile
**
** Hash the contents of the specified file, read in binary mode in
** chunks of up to MAX_BUF_SIZE bytes. Write the result into hashBuf.
*/

void
hashFile(char *hashBuf, char *fileName)
{
    SHA_INFO sha;
    FILE *fp = NULL;
    unsigned char buf[MAX_BUF_SIZE];
    size_t num;


    if (strcmp(fileName, "-") == 0)
    {
	fp = stdin;
#if defined(WIN32) || defined(__CYGWIN__)
	setmode(0, O_BINARY);
#endif
    }
    else if ((fp = fopen(fileName, "rb")) == NULL)
    {
	message(0, errno, "can't open '%s'", fileName);
	return;
    }

    sha_init(&sha);

    while ((num = fread(buf, 1, MAX_BUF_SIZE, fp)) > 0)
    {
	sha_update(&sha, (SHA_BYTE *)buf, num);
    }

    fclose(fp);

    sha_final(&sha);
    sprintf(hashBuf, "%08lx%08lx%08lx%08lx%08lx", sha.digest[0], sha.digest[1],
	    sha.digest[2], sha.digest[3], sha.digest[4]);
}

/*
** checkIdentity
**
** Check the supplied public key string against an "identity file". This
** file contains lines that consist of the SHA hash of the (generator,
** modulus, public-key) tuple followed by some optional commentary text.
** The hash value must begin the line and not contain any white-space.
*/

int
checkIdentity(char *idFile, char *generator, char *modulus, char *key)
{
    FILE *fp;
    char keySig[HASH_STR_SIZE];
    char checkSig[MAX_LINE_SIZE];
    char line[MAX_LINE_SIZE];
    int found = 0;
    int len = 0;


    if ((fp = fopen(idFile, "r")) == NULL)
    {
	message(0, errno, "can't open identity file '%s'", idFile);
	return 0;
    }

    if (*generator == '\0') generator = DFLT_GENERATOR;
    if (*modulus == '\0') modulus = DFLT_MODULUS;

    hashStrings(keySig, generator, modulus, key, NULL);
    len = strlen(keySig);

    message(3, 0, "checking key with identity hash '%s'", keySig);

    while (fgets(line, MAX_LINE_SIZE, fp) != NULL)
    {
	if (sscanf(line, "%s", checkSig) != 1)
	{
	    continue;
	}

	if (strcasecmp(checkSig, keySig) == 0)
	{
	    message(1, 0, "key identity matched: %.*s", strlen(line) - 1, line);
	    found = 1;
	    break;
	}
    }

    fclose(fp);

    return found;
}

/*
** generateIdentity
**
** Given a generator, modulus and private key (exponent) generate
** the hash of the public key string -- the identity used by the
** checkIdentity routine. This is the hash of the (generator,
** modulus, public-key) tuple.
*/

char *
generateIdentity(char *generator, char *modulus, char *exponent)
{
    char *dhKey = diffieHellman(generator, modulus, exponent);
    static char buffer[HASH_STR_SIZE];

    if (*generator == '\0') generator = DFLT_GENERATOR;
    if (*modulus == '\0') modulus = DFLT_MODULUS;

    hashStrings(buffer, generator, modulus, dhKey, NULL);
    free(dhKey);

    return buffer;
}

/*
** makeDetached
**
** Detach the current process from its controlling terminal and run
** in the background. On UNIX this means running as a daemon. On Windows
** this can not completely detach from a parent process that is waiting
** for it but can free itself from the console. To get the full
** effect of a UNIX daemon you should probably also run this as a
** service (see the -S option).
*/

void
makeDetached(void)
{
#ifdef WIN32
    /* Detach from the console */

    if (!FreeConsole())
    {
	message(0, errno, "failed to detach from console (Win32 error = %ld)", GetLastError());
    }

    /* Close stdio streams because they now have nowhere to go! */

    fclose(stdin);
    fclose(stdout);
    fclose(stderr);
#else
    switch (fork())
    {
    case -1:
	/* Error! */
	message(0, errno, "fork failed becoming daemon");
	break;

    case 0:
	/* Child -- continue */
	break;

    default:
	/* Parent -- exit and leave child running */
	exit(EXIT_SUCCESS);
	break;
    }

    /* Detach from controlling terminal */

    setsid();
    close(0);
    close(1);
    close(2);
#endif
}

/*
** allowRedirect
**
** Decide whether redirection to the specified port is allowed or not.
** Returns 1 if it is or 0 otherwise.
*/

int
allowRedirect(unsigned short port)
{
    PortList_t *lp;


    /* If nothing else has been specified then allow any port */

    if (AllowedPorts == NULL)
    {
	return 1;
    }

    /*
    ** Otherwise search the AllowedPorts list to determine wherther the
    ** port lies within any of the permitted ranges.
    */

    for (lp = AllowedPorts; lp; lp = lp->next)
    {
	message(4, 0, "checking port %hu against range %hu - %hu", port, lp->lo, lp->hi);

	if (port >= lp->lo && port <= lp->hi)
	{
	    return 1;
	}
    }

    return 0;
}

/******************************************\
**					  **
**  Core Client/Server Protocol Routines  **
**					  **
\******************************************/


/*
** spawnHandler
**
** This routine creates a single process/thread running the specified
** handler routine to handle the traffic on a single connection.
**
** Note that casting the integer clientFd to void* to pass it to the
** thread creation routine is a little naughty but it works anywhere
** I've tested it!
*/

void
spawnHandler(void (*handler)(int), int listenFd, int clientFd, int inLine)
{
    if (inLine)
    {
	message(4, 0, "running handler function in-line");
	(*handler)(clientFd);
	return;
    }

#if defined(HAVE_PTHREADS)
    {
	pthread_t tid;

	message(4, 0, "spawning handler function thread");
	if (pthread_create(&tid,
			   &ThreadAttr,
			   (void * (*)(void *))handler,
			   (void *)clientFd) == -1)
	{
	    message(0, errno, "failed to create handler thread");
	}
	message(4, 0, "handler thread %lu created", (unsigned long)tid);
    }
#elif defined(WIN32)
    {
	message(4, 0, "spawning handler function thread");
	if (_beginthread((void (*)(void *))handler, THREAD_STACK_SIZE,
			 (LPVOID)clientFd) == 0)
	{
	    message(0, errno, "failed to create handler thread");
	}
	message(4, 0, "handler thread created");
    }
#else	/* No PTHREADS and not WIN32 */
    {
	pid_t pid;


	message(4, 0, "spawning handler sub-process");
	if ((pid = fork()) < 0)
	{
	    message(0, errno, "failed to fork handler sub-process");
	}
        else if (pid == 0)
	{
	    /* Child -- listenFd not needed */

	    closesocket(listenFd);

	    (*handler)(clientFd);
	    exit(EXIT_SUCCESS);
	}
	else
	{
	    /* Parent -- clientFd and args not needed */
	    close(clientFd);
	    message(4, 0, "handler sub-process %lu created", (unsigned long)pid);
	}
    }
#endif
}

/*
** clientListener
**
** This is the top-level client routine that sets up a local socket
** and listens for connections. It the then spawns an individual
** process or thread the handle that single client.
*/

void
clientListener(unsigned short localPort)
{
    int listenFd;
    int clientFd;
    struct sockaddr_in addr;
    int addrLen;


    message(3, 0, "client listener routine entered");

    /*
    ** Create the local listener socket and fire up the requested
    ** sub-command if necessary. We do this first so that the child
    ** process does not inherit the connection to the server.
    */

    message(3, 0, "creating local listener socket");

    if ((listenFd = makeListener(&localPort, 1)) == -1)
    {
	message(0, errno, "can't create listener socket");
	exit(EXIT_FAILURE);
    }

    message(3, 0, "listening on %hu", localPort);

    /* Spawn the sub-command, if specified */

    if (CommandString)
    {
	message(3, 0, "spawning command '%s'", CommandString);

	if (!spawnCommand(localPort, CommandString))
	{
	    exit(EXIT_FAILURE);
	}
    }
    else
    {
	message(1, 0, "Listening on local port %hu", localPort);
    }

    /* Detach from terminal, if required */

    if (IsDetached)
    {
	message(3, 0, "detaching from terminal");
	makeDetached();
    }

    /*
    ** Now wait for a connection from a local client. If we are operating
    ** in "persistent" mode then we will loop forever otherwise this
    ** will be a "one shot" connection.
    */

    do
    {
	message(1, 0, "waiting for client connection on port %hu", localPort);

	addrLen = sizeof(struct sockaddr_in);
	if ((clientFd = accept(listenFd, (struct sockaddr *)&addr, &addrLen)) < 0)
	{
	    message(0, errno, "error on accept");
	}
	else
	{
	    message(1, 0, "accepted connection");
	    spawnHandler(client, listenFd, clientFd, (Debug || !MultiUse));
	}
    }
    while (MultiUse);

    /* We do not need to listen for any more clients */

    closesocket(listenFd);
    listenFd = -1;
}

/*
** client
**
** This routine implements the client side of the Zebedee protocol. It is
** fully re-entrant.
*/

void
client(int clientFd)
{
    const char *serverHost = RemoteHost;
    const unsigned short serverPort = ServerPort;
    const unsigned short redirectPort = RemotePort;
    int serverFd = -1;
    unsigned short response = 0;
    unsigned short size = 0;
    char generator[MAX_LINE_SIZE];
    char modulus[MAX_LINE_SIZE];
    char serverDhKey[MAX_LINE_SIZE];
    char *exponent = NULL;
    char *dhKey = NULL;
    char *secretKeyStr = NULL;
    int len = -1;
    BFState_t *sendState = NULL;
    BFState_t *recvState = NULL;
    unsigned short cmpInfo = CompressInfo;
    unsigned short keyBits = KeyLengthBits;
    unsigned char challenge[CHALLENGE_SIZE];
    unsigned char myChallenge[CHALLENGE_SIZE];
    unsigned short protocol = PROTOCOL_VERSION;


    assert(serverHost != NULL);
    assert(serverPort != 0);
    assert(redirectPort != 0);

    message(3, 0, "client routine entered");

    message(3, 0, "making connection to %s:%hu", serverHost, serverPort);

    if ((serverFd = makeConnection(serverHost, serverPort)) == -1)
    {
	message(0, errno, "can't connect to %s port %hu", serverHost, serverPort);
	goto fatal;
    }
    message(3, 0, "connected to %s:%hu", serverHost, serverPort);

    /*
    ** Request protocol version.
    */

    message(3, 0, "requesting protocol version %#hx", protocol);

    if (!requestResponse(serverFd, protocol, &response))
    {
	message(0, errno, "failed requesting protocol version");
	goto fatal;
    }

    if (response != protocol)
    {
	message(0, 0, "server responded with incompatible protocol version (%#hx)", response);
	goto fatal;
    }
    message(3, 0, "accepted protocol version %#hx", response);

    /* Request compression type and level */

    message(3, 0, "requesting compression level %#hx", CompressInfo);

    if (!requestResponse(serverFd, CompressInfo, &response))
    {
	message(0, errno, "failed requesting compression level");
	goto fatal;
    }

    if (response > CompressInfo)
    {
	message(0, 0, "server responded with invalid compression level (%hu > %hu)", response, cmpInfo);
	goto fatal;
    }
    
    /* We will use only what the server supports */

    cmpInfo = response;
    message(3, 0, "accepted compression level %#hx", response);

    /* Now request redirection to a specific port */

    message(3, 0, "requesting redirection to port %hu", redirectPort);

    if (!requestResponse(serverFd, redirectPort, &response))
    {
	message(0, errno, "failed requesting redirection port");
	goto fatal;
    }

    if (response != redirectPort)
    {
	message(0, errno, "server refused request for redirection to port %hu", redirectPort);
	goto fatal;
    }

    message(3, 0, "redirection request accepted");

    /* Request the key length */

    message(3, 0, "requesting key length %hu", KeyLengthBits);

    if (!requestResponse(serverFd, KeyLengthBits, &response))
    {
	message(0, errno, "failed requesting key length");
	goto fatal;
    }
    keyBits = response;
    message(3, 0, "accepted key length %hu", response);


    /* Read the DH generator */

    message(3, 0, "reading DH generator");

    size = MAX_LINE_SIZE;
    if (readMessage(serverFd, (unsigned char *)generator, &size, cmpInfo, NULL) <= 0)
    {
	message(0, errno, "failed reading DH generator");
	goto fatal;
    }
    message(3, 0, "accepted generator '%s'", generator);


    /* Read the DH modulus */

    message(3, 0, "reading DH modulus");

    size = MAX_LINE_SIZE;
    if (readMessage(serverFd, modulus, &size, cmpInfo, NULL) <= 0)
    {
	message(0, errno, "failed reading DH modulus");
	goto fatal;
    }
    message(3, 0, "accepted modulus '%s'", modulus);


    /* Read the server DH key */

    message(3, 0, "reading server DH key");

    size = MAX_LINE_SIZE;
    if (readMessage(serverFd, serverDhKey, &size, cmpInfo, NULL) <= 0)
    {
	message(0, errno, "failed reading server DH key");
	goto fatal;
    }
    message(3, 0, "accepted server DH key '%s'", serverDhKey);

    /*
    ** If requested, check this against a list of "known" keys
    ** to validate the server identity.
    */

    if (IdentityFile)
    {
	message(3, 0, "checking key against identity file '%s'", IdentityFile);

	if (!checkIdentity(IdentityFile, generator, modulus, serverDhKey))
	{
	    message(0, 0, "server's identity not found in '%s'", IdentityFile);
	    goto fatal;
	}
    }

    /*
    ** Now generate our exponent (the private key). This is returned
    ** as a hex string.
    */

    message(3, 0, "generating private key");

    if ((exponent = generateKey()) == NULL)
    {
	message(0, 0, "can't generate private key");
	goto fatal;
    }
    message(3, 0, "private key generated");


    /*
    ** Generate the public DH key.
    ** If a private key string has been specified then use this.
    */

    message(3, 0, "generating public DH key");

    if ((dhKey = diffieHellman(generator, modulus, exponent)) == NULL)
    {
	message(0, 0, "can't generate public DH key");
	goto fatal;
    }
    message(3, 0, "public DH key is '%s'", dhKey);

    /* Now send the DH key */

    message(3, 0, "sending public DH key");

    assert((len = strlen(dhKey) + 1) <= MAX_LINE_SIZE);
    if (writeMessage(serverFd, (unsigned char *)dhKey,
		     (unsigned short)len, cmpInfo, NULL) != len)
    {
	message(0, errno, "failed writing DH key to server");
	goto fatal;
    }
    message(3, 0, "sent public DH key");

    /* Now generate the shared secret key */

    message(3, 0, "generating shared secret key");

    secretKeyStr = diffieHellman(serverDhKey, modulus, exponent);

    message(3, 0, "shared key ends '...%s'", secretKeyStr + strlen(secretKeyStr) - 4);

    message(3, 0, "initialising encryption state");

    sendState = setupBlowfish(secretKeyStr, keyBits);
    recvState = setupBlowfish(secretKeyStr, keyBits);

    /* Clear unneeded values */

    memset(exponent, 0, strlen(exponent));
    free(exponent);
    exponent = NULL;
    free(dhKey);
    dhKey = NULL;
    memset(secretKeyStr, 0, strlen(secretKeyStr));
    free(secretKeyStr);
    secretKeyStr = NULL;

    /*
    ** Having established the encrypted channel we now perform a mutual
    ** challenge/response dialogue in order to guard against being spoofed
    ** by a replay attack if we are using a static key.
    */

    /* Read encrypted challenge string from the server */

    message(3, 0, "reading challenge from server");

    size = CHALLENGE_SIZE;
    if (readMessage(serverFd, challenge, &size, cmpInfo, recvState) <= 0)
    {
	message(0, errno, "failed to read challenge from server");
	goto fatal;
    }
    message(3, 0, "read challenge");

    /* Calculate the answer and then send that back to the server */

    challengeAnswer(challenge);

    message(3, 0, "writing challenge response");
    if (writeMessage(serverFd, challenge, CHALLENGE_SIZE,
		     cmpInfo, sendState) != CHALLENGE_SIZE)
    {
	message(0, errno, "failed writing challenge response to server");
	goto fatal;
    }
    message(3, 0, "wrote challenge response");

    /* Now generate our own challenge value and send it to the server */

    makeChallenge(myChallenge);

    message(3, 0, "sending challenge to server");

    if (writeMessage(serverFd, myChallenge, CHALLENGE_SIZE,
		     cmpInfo, sendState) != CHALLENGE_SIZE)
    {
	message(0, errno, "failed writing challenge to server");
	goto fatal;
    }
    message(3, 0, "wrote challenge");

    message(3, 0, "reading challenge response from server");

    size = CHALLENGE_SIZE;
    if (readMessage(serverFd, challenge, &size, cmpInfo, recvState) <= 0)
    {
	message(0, errno, "failed to read challenge response from server");
	goto fatal;
    }
    message(3, 0, "read challenge response");

    /* Calculate the expected answer and then compare with the response */

    challengeAnswer(myChallenge);

    if (memcmp(challenge, myChallenge, CHALLENGE_SIZE) != 0)
    {
	message(0, 0, "server responded incorrectly to challenge");
	goto fatal;
    }

    memset(challenge, 0, CHALLENGE_SIZE);
    memset(myChallenge, 0, CHALLENGE_SIZE);

    message(1, 0, "tunnel established to %s port %hu", serverHost, redirectPort);
    message(2, 0, "compression level %#hx, key length %hu", cmpInfo, keyBits);

    /* Now loop handling i/o */

    message(3, 0, "entering filter loop");

    if (!filterLoop(clientFd, serverFd, sendState, recvState, cmpInfo))
    {
	message(0, errno, "failed communicating with server");
	goto fatal;
    }

    message(1, 0, "connection closed");

    closesocket(clientFd);
    closesocket(serverFd);
    if (sendState) free(sendState);
    if (recvState) free(recvState);
#ifdef WIN32
    if (!Debug) _endthread();
#endif
    return;

fatal:
    if (serverFd != -1) closesocket(serverFd);
    if (clientFd != -1) closesocket(clientFd);
    if (exponent) free(exponent);
    if (dhKey) free(dhKey);
    if (secretKeyStr) free(secretKeyStr);
    if (sendState) free(sendState);
    if (recvState) free(recvState);
#ifdef WIN32
    if (!Debug) _endthread();
#endif
    return;
}

/*
** serverListener
**
** This is top-level server routine that listens for incoming connections
** and then spawns an individual process/thread the handle that single
** client.
*/

void
serverListener(unsigned short port)
{
    int listenFd;
    int clientFd;
    struct sockaddr_in addr;
    int addrLen;


    /* Create the listener socket */

    listenFd = makeListener(&port, 0);

    /* Detach from terminal, if requested */

    if (IsDetached)
    {
	message(3, 0, "detaching from terminal");
	makeDetached();
    }

    while (1)
    {
	message(1, 0, "waiting for connection on port %hu", port);

	addrLen = sizeof(struct sockaddr_in);
	if ((clientFd = accept(listenFd, (struct sockaddr *)&addr, &addrLen)) < 0)
	{
	    message(0, errno, "error on accept");
	}
	else
	{
	    message(1, 0, "accepted connection from %s", inet_ntoa(addr.sin_addr));
	    spawnHandler(server, listenFd, clientFd, Debug);
	}
    }
}

/*
** server
**
** This is the server side of the Zebedee protocol. It should match the
** interactions in client(). Note that this routine IS thread-safe and
** should not call exit() directly.
*/

void
server(int clientFd)
{
    int localFd = -1;
    unsigned short request = 0;
    unsigned short response = 0;
    unsigned short cmpInfo = CompressInfo;
    unsigned short keyBits = KeyLengthBits;
    unsigned short size;
    unsigned short port;
    char clientDhKey[MAX_LINE_SIZE];
    char *exponent = NULL;
    char *dhKey = NULL;
    char *secretKeyStr = NULL;
    int len = -1;
    BFState_t *sendState = NULL;
    BFState_t *recvState = NULL;
    unsigned char challenge[CHALLENGE_SIZE];
    unsigned char myChallenge[CHALLENGE_SIZE];


    message(3, 0, "server routine entered");

    /* Read protocol version */

    message(3, 0, "reading protocol version ");
    
    if (readUShort(clientFd, &request) != 2)
    {
	message(0, errno, "failed reading protocol version");
	goto fatal;
    }

    message(3, 0, "read protocol version %#hx", request);

    /*
    ** If the client protocol version matches one that we can support
    ** then we just echo it back otherwise a response of zero means that
    ** the server cannot support the version requested by the client.
    **
    ** Currently there is only one version so this is an easy decision ...
    */

    if (request == PROTOCOL_VERSION)
    {
	response = request;
    }
    else
    {
	message(0, 0, "client requested unsupported protocol version %#hx", request);
	response = 0;
    }

    message(3, 0, "replying with protocol version %#hx", response);
    
    if (writeUShort(clientFd, response) != 2)
    {
	message(0, errno, "failed writing protocol version back to client");
	goto fatal;
    }

    /* Read compression level */

    message(3, 0, "reading compression level");

    if (readUShort(clientFd, &request) != 2)
    {
	message(0, errno, "failed reading compression level");
	goto fatal;
    }
    message(3, 0, "read compression level %#hx", request);

    /*
    ** Use the minimum of the client's and server's compression levels.
    **
    ** Note that all values for zlib compression are less than those for
    ** bzip2 compression so if a client requests bzip2 compression but
    ** the server doesn't support it then the protocol degrades naturally
    ** to zlib.
    */

    if (request < cmpInfo)
    {
	cmpInfo = request;
	response = request;
    }
    else
    {
	response = cmpInfo;
    }

    message(3, 0, "replying with compression level %#hx", response);

    if (writeUShort(clientFd, response) != 2)
    {
	message(0, errno, "failed writing compression level back to client");
	goto fatal;
    }

    /* Now read port redirection request */

    message(3, 0, "reading redirection port");

    if (readUShort(clientFd, &request) != 2)
    {
	message(0, errno, "failed reading redirection port");
	goto fatal;
    }
    message(3, 0, "read port %hu", request);

    /*
    ** The server should not, in general, redirect arbitrary ports because
    ** the remote client will appear to the "target" service to have
    ** connected from the local machine -- and may assume that greater
    ** access should be allowed as a result. So we check the requested
    ** port against the list of allowed ports. If the request is not
    ** granted then we send zero back to the client otherwise we carry
    ** on and attempt to open a connection to that port on the local
    ** host. If that succeeds we will send back the requested port number
    ** or zero if it fails.
    */

    message(3, 0, "checking if redirection is allowed");

    if (!allowRedirect(request))
    {
	message(0, 0, "client requested redirection to a disallowed port (%hu)", request);
	writeUShort(clientFd, 0);
	goto fatal;
    }

    message(3, 0, "allowed redirection request to %hu", request);

    /*
    ** Attempt to open connection -- if this fails then we write back
    ** zero to the client otherwise we send back the requested port
    ** number.
    */

    message(3, 0, "opening connection to port %hu on localhost", request);

    if ((localFd = makeConnection("localhost", request)) == -1)
    {
	message(0, errno, "failed connecting to port %hu", request);
	writeUShort(clientFd, 0);
	goto fatal;
    }
    port = request;

    message(3, 0, "made local connection -- writing back %hu to client", request);

    /* All OK -- echo back port number */

    if (writeUShort(clientFd, request) != 2)
    {
	message(0, errno, "failed writing redirection port back to client");
	goto fatal;
    }

    /* Read key length request */

    message(3, 0, "reading key length");

    if (readUShort(clientFd, &request) != 2)
    {
	message(0, errno, "failed reading key length");
	goto fatal;
    }

    message(3, 0, "read key length %hu", request);

    /* Use the minimum of the client and server's values */

    if (request > keyBits)
    {
	response = keyBits;
    }
    else
    {
	keyBits = request;
	response = request;
    }

    message(3, 0, "replying with key length %hu", response);

    if (writeUShort(clientFd, response) != 2)
    {
	message(0, errno, "failed writing key length back to client");
	goto fatal;
    }

    /*
    ** Send the Diffie-Hellman generator
    **
    ** The server decides the generator and modulus because to avoid
    ** "man-in-the-middle" attacke you are more likely to want to know
    ** the identity of the server from the client than vice-versa. This
    ** lets the server control its public key identity better.
    */

    message(3, 0, "sending DH generator '%s'", Generator);

    assert((len = strlen(Generator) + 1) <= MAX_LINE_SIZE);
    if (writeMessage(clientFd, (unsigned char *)Generator,
		     (unsigned short)len, cmpInfo, NULL) != len)
    {
	message(0, errno, "failed writing generator to client");
	goto fatal;
    }

    message(3, 0, "sent generator");

    /* Send the Diffie-Hellman modulus */

    message(3, 0, "sending DH modulus '%s'", Modulus);

    assert((len = strlen(Modulus) + 1) <= MAX_LINE_SIZE);
    if (writeMessage(clientFd, (unsigned char *)Modulus,
		     (unsigned short)len, cmpInfo, NULL) != len)
    {
	message(0, errno, "failed writing modulus to client");
	goto fatal;
    }

    message(3, 0, "sent modulus", Modulus);


    /*
    ** Calculate our DH key
    **
    ** This requires that we first generate our private key (the
    ** the exponent) and then perform the modular exponentiation
    ** to generate the public key.
    */

    message(3, 0, "generating private key");

    if ((exponent = generateKey()) == NULL)
    {
	message(0, errno, "can't generate private key");
	goto fatal;
    }

    message(3, 0, "private key generated");

    message(3, 0, "generating public DH key");

    if ((dhKey = diffieHellman(Generator, Modulus, exponent)) == NULL)
    {
	message(0, errno, "can't generate public DH key");
	goto fatal;
    }

    message(3, 0, "public DH key is '%s'", dhKey);

    /* Now send the DH key */

    message(3, 0, "sending public DH key");

    assert((len = strlen(dhKey) + 1) <= MAX_LINE_SIZE);
    if (writeMessage(clientFd, (unsigned char *)dhKey,
		     (unsigned short)len, cmpInfo, NULL) != len)
    {
	message(0, errno, "failed writing DH key to client");
	goto fatal;
    }

    message(3, 0, "sent public DH key");


    /* Read the client DH key */

    message(3, 0, "reading client DH key");

    size = MAX_LINE_SIZE;
    if (readMessage(clientFd, clientDhKey, &size, cmpInfo, NULL) <= 0)
    {
	message(0, errno, "failed reading client DH key");
	goto fatal;
    }

    message(3, 0, "accepted client DH key '%s'", clientDhKey);


    /*
    ** If requested, check this against a list of "known" keys
    ** to validate the client identity.
    */

    if (IdentityFile)
    {
	message(3, 0, "checking key against identity file '%s'", IdentityFile);

	if (!checkIdentity(IdentityFile, Generator, Modulus, clientDhKey))
	{
	    message(0, 0, "client's key identity not found in '%s'", IdentityFile);
	    goto fatal;
	}
    }


    /* Now generate the shared secret key */

    message(3, 0, "generating shared secret key");

    secretKeyStr = diffieHellman(clientDhKey, Modulus, exponent);

    message(3, 0, "shared key ends '...%s'", secretKeyStr + strlen(secretKeyStr) - 4);

    message(3, 0, "initialising encryption state");

    sendState = setupBlowfish(secretKeyStr, keyBits);
    recvState = setupBlowfish(secretKeyStr, keyBits);

    /* Clear unneeded values */

    free(dhKey);
    dhKey = NULL;
    memset(exponent, 0, strlen(exponent));
    free(exponent);
    exponent = NULL;
    memset(secretKeyStr, 0, strlen(secretKeyStr));
    free(secretKeyStr);
    secretKeyStr = NULL;

    /*
    ** Having established the encrypted channel we now perform a mutual
    ** challenge/response dialogue in order to guard against being spoofed
    ** by a replay attack if we are using a static key.
    */

    /* Generate our own challenge value */

    makeChallenge(myChallenge);

    message(3, 0, "sending challenge to client");

    if (writeMessage(clientFd, myChallenge, CHALLENGE_SIZE,
		     cmpInfo, sendState) != CHALLENGE_SIZE)
    {
	message(0, errno, "failed writing challenge to client");
	goto fatal;
    }
    message(3, 0, "wrote challenge");

    message(3, 0, "reading challenge response from client");

    size = CHALLENGE_SIZE;
    if (readMessage(clientFd, challenge, &size, cmpInfo, recvState) <= 0)
    {
	message(0, errno, "failed to read challenge response from client");
	goto fatal;
    }
    message(3, 0, "read challenge response");

    /* Calculate the expected answer and then compare with the response */

    challengeAnswer(myChallenge);

    if (memcmp(challenge, myChallenge, CHALLENGE_SIZE) != 0)
    {
	message(0, 0, "client responded incorrectly to challenge");
	goto fatal;
    }

    /* Now read challenge from the client */

    message(3, 0, "reading challenge from client");

    size = CHALLENGE_SIZE;
    if (readMessage(clientFd, challenge, &size, cmpInfo, recvState) <= 0)
    {
	message(0, errno, "failed to read challenge from client");
	goto fatal;
    }
    message(3, 0, "read challenge");

    /* Calculate the answer and then send that back to the client */

    challengeAnswer(challenge);

    message(3, 0, "writing challenge response");

    if (writeMessage(clientFd, challenge, CHALLENGE_SIZE,
		     cmpInfo, sendState) != CHALLENGE_SIZE)
    {
	message(0, errno, "failed writing challenge response to client");
	goto fatal;
    }
    message(3, 0, "wrote challenge response");
    memset(challenge, 0, CHALLENGE_SIZE);

    message(1, 0, "tunnel established to port %hu", port);
    message(2, 0, "compression level %#hx, key length %hu", cmpInfo, keyBits);

    /* Now loop handling i/o */

    message(3, 0, "entering filter loop");

    if (!filterLoop(localFd, clientFd, sendState, recvState, cmpInfo))
    {
	message(0, errno, "failed communicating with client");
	goto fatal;
    }

    errno = 0;
    message(1, 0, "connection closed");

    closesocket(clientFd);
    closesocket(localFd);
    if (sendState) free(sendState);
    if (recvState) free(recvState);
#ifdef WIN32
    if (!Debug) _endthread();
#endif
    return;

fatal:
    if (clientFd != -1) closesocket(clientFd);
    if (localFd != -1) closesocket(localFd);
    if (exponent) free(exponent);
    if (dhKey) free(dhKey);
    if (secretKeyStr) free(secretKeyStr);
    if (sendState) free(sendState);
    if (recvState) free(recvState);
#ifdef WIN32
    if (!Debug) _endthread();
#endif
}

/*****************************\
**			     **
**  Option Parsing Routines  **
**			     **
\*****************************/

/*
** scanPortRange
**
** Parse a port range specifier and place the high and low ends of the
** range in hiP and loP. A range can be two shorts integers separated by
** a "-" -- for example "5900-5910", a single number or a symbolic port name.
** In the latter two cases both hiP and loP are set the same.
**
** If either loP or hiP are NULL the corresponding value is not set.
** The command always returns the vale of the low port.
*/

unsigned short
scanPortRange(const char *str, unsigned short *loP, unsigned short *hiP)
{
    struct servent *entry = NULL;
    unsigned short loVal = 0;
    unsigned short hiVal = 0;
    char portName[MAX_LINE_SIZE];


    switch (sscanf(str, "%hu-%hu", &loVal, &hiVal))
    {
    case 0:
	break;

    case 1:
	hiVal = loVal;

	/* Fall through */

    case 2:
	if (hiVal < loVal)
	{
	    hiVal ^= loVal;
	    hiVal = hiVal ^ loVal;
	    hiVal ^= loVal;
	}

	if (loVal < 1 || hiVal > 65535)
	{
	    message(0, 0, "invalid port range '%s'", str);
	    return 0;
	}

	if (hiVal != loVal && !hiP)
	{
	    message(0, 0, "port range found where single value expected");
	    return 0;
	}

	if (loP) *loP = loVal;
	if (hiP) *hiP = hiVal;

	return loVal;
    }

    if (sscanf(str, "%s", portName) != 1)
    {
	message(0, 0, "missing port name");
	return 0;
    }

    if ((entry = getservbyname(portName, "tcp")) == NULL)
    {
	message(0, errno, "can't find port name entry for '%s'", portName);
	return 0;
    }

    loVal = ntohs(entry->s_port);
    if (loP) *loP = loVal;
    if (hiP) *hiP = loVal;

    return loVal;
}

/*
** setBoolean
**
** Parse a boolean value (either "true" or "false") and set the supplied
** result appropriately.
*/

void
setBoolean(char *value, int *resultP)
{
    if (strcasecmp(value, "true") == 0)
    {
	*resultP = 1;
    }
    else if (strcasecmp(value, "false") == 0)
    {
	*resultP = 0;
    }
    else
    {
	message(0, 0, "can't parse boolean value '%s'", value);
    }
}

/*
** setUShort
**
** Parse and set an unsigned short value
*/

void
setUShort(char *value, unsigned short *resultP)
{
    if (sscanf(value, "%hu", resultP) != 1)
    {
	message(0, 0, "can't parse unsigned short value '%s'", value);
    }
}

/*
** setPort
**
** Parse a port name or number and set an unsigned short value
*/

void
setPort(char *value, unsigned short *resultP)
{
    unsigned short port;


    port = scanPortRange(value, NULL, NULL);
    if (port == 0)
    {
	message(0, 0, "can't parse port value '%s' key", value);
    }
    else
    {
	*resultP = port;
    }
}

/*
** setAllowedPorts
**
** Parse a list of white-space separated ports or port ranges and create
** the AllowedPorts list. Each element of this list is a low-high pair
** of port numbers. Where single ports are specified both low and high
** are the same.
*/

void
setAllowedPorts(char *value)
{
    PortList_t *new = NULL;
    char *token = NULL;


    token = value;
    while (*token)
    {
	/* Skip whitespace */

	while (*token && isspace(*token)) token++;
	if (!*token) break;

	/* Allocate new list element */

	if ((new = (PortList_t *)malloc(sizeof(PortList_t))) == NULL)
	{
	    message(0, errno, "failed allocating memory for port list");
	    exit(EXIT_FAILURE);
	}

	/* Get high and low port values from the token */

	if (scanPortRange(token, &(new->lo), &(new->hi)) != 0)
	{
	    new->next = AllowedPorts;
	    AllowedPorts = new;
	}

	/* Skip to the next gap */

	while (*token && !isspace(*token)) token++;
    }
}

/*
** setString
**
** Set a character string value
*/

void
setString(char *value, char **resultP)
{
    if ((*resultP = (char *)malloc(strlen(value) + 1)) == NULL)
    {
	message(0, errno, "failed allocating space for string value '%s'", value);
	exit(EXIT_FAILURE);
    }

    strcpy(*resultP, value);
}

/*
** setLogFile
**
** Open the specified log-file for append, closing any currently open
** log.
*/

void
setLogFile(char *newFile)
{
    if (LogFileP != NULL)
    {
	fclose(LogFileP);
    }

    if ((LogFileP = fopen(newFile, "a")) == NULL)
    {
	message(0, errno, "can't open log file '%s'", newFile);
    }
}

/*
** setCmpInfo
**
** Set compression type and level
*/

void
setCmpInfo(char *value, unsigned short *resultP)
{
    unsigned short level = 0;
    unsigned short type = CMPTYPE_ZLIB;

    if (sscanf(value, "zlib:%hu", &level) == 1)
    {
	type = CMPTYPE_ZLIB;
    }
    else if (sscanf(value, "bzip2:%hu", &level) == 1)
    {
#ifdef DONT_HAVE_BZIP2
	message(1, 0, "Warning: bzip2 compression not supported -- using zlib")
	type = CMPTYPE_ZLIB;
#else
	type = CMPTYPE_BZIP2;

	/* Compression level > 1 is useless as data buffers are too small */

	if (level > 1) level = 1;
#endif
    }
    else if (sscanf(value, "%hu", &level) == 1)
    {
	type = CMPTYPE_ZLIB;
    }
    else
    {
	message(0, 0, "invalid compression specification '%s'", value);
	level = DFLT_CMP_LEVEL;
    }

    if (level > 9)
    {
	message(0, 0, "compression level out of range (%s)", value);
	level = DFLT_CMP_LEVEL;
    }

    if (level == 0)
    {
	type = CMPTYPE_ZLIB;
    }

    *resultP = level;
    *resultP = SET_CMPTYPE(*resultP, type);
}

/*
** readConfigFile
**
** Read a configuration file. If the level is greater than MAX_LEVEL the
** recursive inclusion has probably been detected.
*/

void
readConfigFile(const char *fileName, int level)
{
    FILE *fp = NULL;
    char lineBuf[MAX_LINE_SIZE];
    char key[MAX_LINE_SIZE];
    char value[MAX_LINE_SIZE];
    char comment[2];
    char *curPtr = NULL;
    int size = 0;
    int len = 0;
    int lineNo = 0;


    message(2, 0, "reading config file '%s' at level %d", fileName, level);

    if (level > MAX_INCLUDE)
    {
	message(0, 0, "include file nesting too deep (> %d)", MAX_INCLUDE);
	return;
    }

    if ((fp = fopen(fileName, "r")) == NULL)
    {
	message(0, errno, "can't open config file '%s'", fileName);
	return;
    }

    curPtr = lineBuf;
    size = MAX_LINE_SIZE;

    while (fgets(curPtr, size, fp) != NULL)
    {
	lineNo++;

	len = strlen(curPtr) - 1;

	/* Strip new-line */

	if (curPtr[len] == '\n')
	{
	    curPtr[len--] = '\0';
	}
	else
	{
	    message(0, 0, "line too long in config file '%s' at line %d", fileName, lineNo);
	    break;
	}

	message(3, 0, "line %d: %s", lineNo, curPtr);

	/* Look for continued lines */

	if (curPtr[len] == '\\')
	{
	    curPtr[len] = '\0';
	    size -= len;
	    curPtr += len;

	    /* Go and read some more */

	    continue;
	}

	/* Reset for next time */

	curPtr = lineBuf;
	size = MAX_LINE_SIZE;

	/* Split into key-value pairs */

	*comment = '#';
	*key = '#';
	if (sscanf(lineBuf, "%s \"%[^\"]\" %1s", key, value, comment) < 2 &&
	    sscanf(lineBuf, "%s \'%[^\']' %1s", key, value, comment) < 2 &&
	    sscanf(lineBuf, "%s %s %1s", key, value, comment) < 2)
	{
	    /* Issue a message if the line was not blank or a comment */

	    if (*key != '#')
	    {
		message(0, 0, "invalid line in config file '%s' at line %d", fileName, lineNo);
	    }
	    continue;
	}

	/* Skip pure comment lines */

	if (*key == '#') continue;
	    
	/* Third field if present must be a comment */

	if (*comment != '#')
	{
	    message(0, 0, "invalid line in config file '%s' at line %d", fileName, lineNo);
	}

	message(3, 0, "key = '%s', value = '%s'", key, value);

	if (!strcasecmp(key, "server")) setBoolean(value, &IsServer);
	else if (!strcasecmp(key, "detached")) setBoolean(value, &IsDetached);
	else if (!strcasecmp(key, "debug")) setBoolean(value, &Debug);
	else if (!strcasecmp(key, "compression")) setCmpInfo(value, &CompressInfo);
	else if (!strcasecmp(key, "keylength")) setUShort(value, &KeyLengthBits);
	else if (!strcasecmp(key, "verbosity")) setUShort(value, &LogLevel);
	else if (!strcasecmp(key, "serverport")) setPort(value, &ServerPort);
	else if (!strcasecmp(key, "localport")) setPort(value, &LocalPort);
	else if (!strcasecmp(key, "remoteport")) setPort(value, &RemotePort);
	else if (!strcasecmp(key, "remotehost")) setString(value, &RemoteHost);
	else if (!strcasecmp(key, "command")) setString(value, &CommandString);
	else if (!strcasecmp(key, "keygencommand")) setString(value, &KeyGenCmd);
	else if (!strcasecmp(key, "logfile")) setLogFile(value);
	else if (!strcasecmp(key, "timestamplog")) setBoolean(value, &TimestampLog);
	else if (!strcasecmp(key, "multiuse")) setBoolean(value, &MultiUse);
	else if (!strcasecmp(key, "include")) readConfigFile(value, level+1);
	else if (!strcasecmp(key, "modulus")) setString(value, &Modulus);
	else if (!strcasecmp(key, "generator")) setString(value, &Generator);
	else if (!strcasecmp(key, "privatekey")) setString(value, &PrivateKey);
	else if (!strcasecmp(key, "checkidfile")) setString(value, &IdentityFile);
	else if (!strcasecmp(key, "redirect")) setAllowedPorts(value);
	else if (!strcasecmp(key, "message")) message(1, 0, "%s", value);
	else if (!strcasecmp(key, "name")) setString(value, &Program);
	else
	{
	    message(0, 0, "invalid line in config file '%s' at line %d", fileName, lineNo);
	}
    }

    fclose(fp);
}

/*
** cleanHexStr
**
** Canonicalize a hexadecimal string converting to lower case and
** eliminating white-space.
*/

char *
cleanHexString(char *str)
{
    char *newStr = (char *)malloc(strlen(str) + 1);
    char *s = newStr;


    while (*str)
    {
	if (isxdigit(*str))
	{
	    *s++ = tolower(*str++);
	}
    }
    *s = '\0';

    return newStr;
}

/*
** usage
**
** Print out usage message and exit
*/

void
usage(void)
{
    fprintf(stderr, "Zebedee -- A Secure Tunnel Program: Release %s\n", RELEASE_STR);
    fprintf(stderr, "Copyright (c) 1999 by Neil Winton. All Rights Reserved.\n");
    fprintf(stderr,
	    "This program is free software and may be distributed under the terms of the\n"
	    "GNU General Public License, Version 2.\n");
    fprintf(stderr, "Zebedee comes with ABSOLUTELY NO WARRANTY.\n\n");
    fprintf(stderr,
	    "Client: %s [options] [[localport:]host[:remoteport]]\n"
	    "Server: %s [options] -s\n"
	    "Key generation:  %s -p | -P [-f file]\n"
	    "Hashing: %s -H string ... | -h [file ...]\n"
#ifdef WIN32
	    "Service: %s [options] -S [install[=config-file] | remove | run]\n",
	    Program
#endif
	    , Program, Program, Program, Program);

    fprintf(stderr,
	    "Options are:\n"
	    "    -D          Debug mode\n"
	    "    -d          Do not detach from terminal\n"
	    "    -e command  Run command connected to local port (client only)\n"
	    "    -f file     Read configuration file\n"
	    "    -H          Generate hash of string values\n"
	    "    -h          Generate hash of file contents\n"
	    "    -k keybits  Specify key length in bits\n"
	    "    -m          Client accepts multiple connections\n"
	    "    -n name     Specify program name\n"
	    "    -o file     Log output to specified file\n"
	    "    -p          Generate private key\n"
	    "    -P          Generate public key \"fingerprint\"\n"
	    "    -r ports    Specify allowed port redirection list (server only)\n"
	    "    -s          Run as a server\n"
#ifdef WIN32
	    "    -S option   Install/remove/run service\n"
#endif
	    "    -t          Timestamp log entries\n"
	    "    -v level    Set message verbosity level (default 1)\n"
	    "    -z type     Set the compression type and level (default zlib:6)\n");

    exit(EXIT_FAILURE);
}

/*
** sigpipeHandler
**
** Signal handler to ensure that SIGPIPE is caught but does nothing except
** interrupt a system call.
*/

void
sigpipeHandler(int sig)
{
}

/******************\
**		  **
**  Main Routine  **
**		  **
\******************/

int
main(int argc, char **argv)
{
    int ch;
    char hostName[MAX_LINE_SIZE];
    int doHash = 0;
    int doPrivKey = 0;
    int doPubKey = 0;
    char hashBuf[HASH_STR_SIZE];
    char *last;
    char *serviceArgs = NULL;
    char localBuf[100];
    char remoteBuf[100];


    /* Set program name to the last element of the path minus extension */

    if ((last = strrchr(argv[0], FILE_SEP_CHAR)) != NULL)
    {
	Program = last + 1;
    }
    else
    {
	Program = argv[0];
    }
    if ((last = strrchr(Program, '.')) != NULL)
    {
	*last = '\0';
    }

    /* Initialise critical stuff */

    threadInit();
#ifdef WIN32
    if (WSAStartup(0x0101, &WsaState) != 0)
    {
	message(0, errno, "winsock initailization failed");
    }
#endif

    /* Parse the options! */

    while ((ch = getopt(argc, argv, "Dde:f:hHk:mn:o:pPr:sS:tv:z:")) != -1)
    {
	switch (ch)
	{
	case 'D':
	    Debug = 1;
	    break;

	case 'd':
	    IsDetached = 0;
	    break;

	case 'e':
	    CommandString = optarg;
	    break;

	case 'f':
	    readConfigFile(optarg, 1);
	    break;

	case 'h':
	    if (doPrivKey || doPubKey)
	    {
		message(0, 0, "-h and -p or -P are mutually exclusive");
		exit(EXIT_FAILURE);
	    }
	    doHash = 1;
	    break;

	case 'H':
	    if (doPrivKey || doPubKey)
	    {
		message(0, 0, "-H and -p or -P are mutually exclusive");
		exit(EXIT_FAILURE);
	    }
	    doHash = 2;
	    break;

	case 'k':
	    setUShort(optarg, &KeyLengthBits);
	    break;

	case 'm':
	    MultiUse++;
	    break;

	case 'n':
	    Program = optarg;
	    break;

	case 'o':
	    setLogFile(optarg);
	    break;

	case 'p':
	    if (doHash)
	    {
		message(0, 0, "-h or -H and -p or -P are mutually exclusive");
		exit(EXIT_FAILURE);
	    }
	    doPrivKey++;
	    break;

	case 'P':
	    if (doHash)
	    {
		message(0, 0, "-h or -H and -p or -P are mutually exclusive");
		exit(EXIT_FAILURE);
	    }
	    doPubKey++;
	    break;

	case 'r':
	    setAllowedPorts(optarg);
	    break;

	case 's':
	    IsServer = 1;
	    break;

	case 'S':
	    serviceArgs = optarg;
	    break;

	case 't':
	    TimestampLog = 1;
	    break;

	case 'v':
	    LogLevel = strtoul(optarg, NULL, 10);
	    break;

	case 'z':
	    setCmpInfo(optarg, &CompressInfo);
	    break;

	default:
	    usage();
	    break;
	}
    }

    if (CommandString && MultiUse)
    {
	message(0, 0, "can't specify a command for a multi-use client");
	exit(EXIT_FAILURE);
    }

    /*
    ** Clean up and canonicalize the modulus and generator strings.
    ** We do this so that slightly different representations of these
    ** values (different case and white space) will be sent across the
    ** wire in the same form. This is important for the identity
    ** checking feature which calculates a hash of these string values.
    */

    Modulus = cleanHexString(Modulus);
    Generator = cleanHexString(Generator);

#ifdef SIGPIPE
    /*
    ** Set up a handler for SIGPIPE so that is will interrupt a select()
    ** call but otherwise do nothing.
    */

    signal(SIGPIPE, sigpipeHandler);
#endif

#ifdef WIN32
    if (serviceArgs)
    {
	int status = EXIT_SUCCESS;


	if (!strncmp(serviceArgs, "install", 7))
	{
	    if (strncmp(serviceArgs, "install=", 8) == 0)
	    {
		status = svcInstall(Program, serviceArgs + 8);
	    }
	    else
	    {
		status = svcInstall(Program, NULL);
	    }
	}
	else if (!strcmp(serviceArgs, "remove"))
	{
	    status = svcRemove(Program);
	}
	else if (!strcmp(serviceArgs, "run"))
	{
	    svcRun(Program);
	}
	else
	{
	    message(0, 0, "invalid argument to -S option: %s", serviceArgs);
	    status = EXIT_FAILURE;
	}

	exit(status);
    }
#endif

    if (doHash)
    {
	if (optind >= argc)
	{
	    if (doHash == 1)
	    {
		hashFile(hashBuf, "-");
		printf("%s\n", hashBuf);
	    }
	    else
	    {
		message(0, 0, "no string argument to hash");
		exit(EXIT_FAILURE);
	    }
	}
	else
	{
	    while (optind < argc)
	    {
		if (doHash == 1)
		{
		    hashFile(hashBuf, argv[optind]);
		}
		else
		{
		    hashStrings(hashBuf, argv[optind], NULL);
		}
		printf("%s %s\n", hashBuf, argv[optind]);
		optind++;
	    }
	}
    }
    else if (doPrivKey || doPubKey)
    {
	if (doPrivKey)
	{
	    PrivateKey = generateKey();
	    if (PrivateKey != NULL)
	    {
		printf("privatekey \"%s\"\n", PrivateKey);
	    }
	    else
	    {
		message(0, errno, "can't generate private key");
	    }
	}
	if (doPubKey)
	{
	    if (PrivateKey == NULL)
	    {
		message(0, 0, "can't generate a identity without a private key being set");
		exit(EXIT_FAILURE);
	    }
	    gethostname(hostName, MAX_LINE_SIZE);
	    printf("%s %s\n", generateIdentity(Generator, Modulus, PrivateKey), hostName);
	}
    }
    else if (IsServer)
    {
	serverListener(ServerPort);
    }
    else
    {
	switch (argc - optind)
	{
	case 1:
	    if (sscanf(argv[optind], "%[^:]:%[^:]:%s", localBuf, hostName, remoteBuf) == 3)
	    {
		LocalPort = scanPortRange(localBuf, NULL, NULL);
		RemoteHost = hostName;
		RemotePort = scanPortRange(remoteBuf, NULL, NULL);
	    }
	    else if (sscanf(argv[optind], "%[^:]:%s", hostName, remoteBuf) == 2)
	    {
		RemoteHost = hostName;
		RemotePort = scanPortRange(remoteBuf, NULL, NULL);
	    }
	    else
	    {
		RemoteHost = argv[optind];
	    }
	    break;

	case 0:
	    if (RemoteHost == NULL)
	    {
		message(0, 0, "no remote host specified");
		exit(EXIT_FAILURE);
	    }
	    break;

	default:
	    usage();
	    break;
	}

	clientListener(LocalPort);
    }

    exit(EXIT_SUCCESS);
}
